class Frequency():
    """Class to generate frequency of initial text
    """
    def __init__(self, text):
        self.text = text
        self.freq = {}

        for character in text:
            if character in self.freq:
                self.freq[character] += 1
            else:
                self.freq[character] = 1

        self.freq = dict(sorted(self.freq.items(), key=lambda x: x[1]))

    def get_text(self):
        """Get initial text

        Returns:
            str: initial text
        """
        return self.text

    def get_freq(self):
        """Get generate frequency

        Returns:
            dict: frequency of each characters in initial text
        """
        return self.freq



class Node():
    """A node is an element with up to 2 children, a leaf has one character.
    """
    def __init__(self, weight, char=None):
        self.left = None
        self.right = None
        self.char = char
        self.weight = weight

    def setChildren(self, left_node, right_node):
        """Define children of Node

        Args:
            left_node (Node): [description]
            right_node (Node): [description]
        """
        self.left = left_node
        self.right = right_node

    def get_left(self):
        return self.left

    def get_right(self):
        return self.right

    def get_char(self):
        return self.char

    def get_weight(self):
        return self.weight

    def __str__(self):
        return f"char : {self.char} - weight : {self.weight}"



from .File import File
from .Tree import Tree
from .Encodage import Encodage
from .Frequency import Frequency
from .Ratio import Ratio
from .HuffmanInterface import HuffmanInterface


class Huffman(HuffmanInterface):
    """Encode data into smaller bits according to their frequencies of occurance.
    """
    def compressFile(self, path: str) -> dict:
        """Overrides HuffmanInterface.compressFile()

        Args:
            path (str): path of the file to compress

        Returns dict:
            Frequency: Frequency of characters in file
            Str: Binary text of the compressed file
            Float: Percentage reduction
            Float: Average number of bits to encode a character
        """
        file = File(path)
        text = file.read()
        compressedData = self.compress(text)
        file.export(
            compressedData['frequency'],
            compressedData['compressedText']
        )
        return compressedData

    def compress(self, text: str) -> dict:
        """Overrides HuffmanInterface.compress()

        Args:
            text (str): text to compress

        Returns dict:
            Frequency: Frequency of characters in file
            Str: Binary text of the compressed file
            Float: Percentage reduction
            Float: Average number of bits to encode a character
        """
        frequency = Frequency(text)
        tree = Tree(frequency.get_freq())
        encodage = Encodage(tree.get_root())
        compressedText = encodage.encode(text)
        ratio = Ratio(frequency.get_freq(), encodage.get_table())
        return {
            'frequency': frequency.get_freq(),
            'compressedText': compressedText,
            'ratio': ratio.getRatio(),
            'averageBitsForChar': ratio.getAveragageBitsForChar()
        }



class Encodage():
    """Class to generate encodage table and use it.
    """
    def __init__(self, node):
        self.table = {}
        self.make_table(node)

    def make_table(self, node, code=""):
        """Using recursivity to generate table for encoding text

        Args:
            node (Node): The node is the root node in the first iteration,
                and after that it is a child node until it reaches a leaf.
            code (str, optional): Defaults to "" in first iteration,
                and is completed by 0 or 1 at each iteration.

        Returns:
            Dict: Binary table for code each characters
        """
        if node.get_char():
            if not code:
                self.table[node.get_char()] = "0"
            else:
                self.table[node.get_char()] = code
            return self.table
        else:
            self.table.update(self.make_table(node.get_left(), code+"0"))
            self.table.update(self.make_table(node.get_right(), code+"1"))
            return self.table

    def get_table(self):
        """Return binary table for code each characters

        Returns:
            Dict: Binary table for code each characters
        """
        return self.table

    def encode(self, text):
        """Encodes the text passed in parameter

        Args:
            text (str): Text to compress

        Returns:
            String: Encoded text
        """
        return "".join([self.get_table()[char] for char in text]).encode()



class Ratio():
    """Get informations (statistics) on compression
    """
    def __init__(self, frequency, encodage_table):
        self.frequency = frequency
        self.encodage_table = encodage_table

    def getRatio(self):
        """Determine the compression ratio

        Returns:
            float: compression ratio
        """
        nbChars = 0
        compressedBits = 0
        for char in self.frequency.items():
            nbChars += char[1]
            compressedBits += char[1]*len(self.encodage_table[char[0]])

        originalBits = nbChars * 8
        return 1 - compressedBits / originalBits

    def getAveragageBitsForChar(self) -> float:
        """Average number of storage bits of a character in the compressed text

        Returns:
            float: average bits for one char
        """
        bitsForChar = []

        for char in self.encodage_table.values():
            bitsForChar.append(len(char))

        return sum(bitsForChar)/len(bitsForChar)



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

setup(
    name='huffman-project',
    version='0.1.0',
    description='huffman-project package to reduce data based on character frequency',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Maverick Perrollaz',
    author_email='maverick.perrollaz@pm.me',
    url='https://github.com/M4verickFr/huffman-project',
    license='Apache 2 License',
    packages=find_packages()
)



from .Node import Node


class Tree():
    def __init__(self, frequency):
        list_nodes = [
            Node(weight, character) for character, weight in frequency.items()]

        while len(list_nodes) > 1:
            list_nodes = sorted(list_nodes, key=lambda x: x.get_weight())
            left_node = list_nodes.pop(0)
            right_node = list_nodes.pop(0)
            parent_node = Node(
                left_node.get_weight() + right_node.get_weight())
            parent_node.setChildren(left_node, right_node)
            list_nodes.append(parent_node)

        self.root = list_nodes[0]

    """Get root Node of Tree

    Returns:
        Node: root node
    """
    def get_root(self):
        return self.root



class HuffmanInterface:
    def compressFile(self, path: str) -> dict:
        """Compress the file provided in parameter.

        Args:
            path (str): path of the file to compress
        """
        pass

    def compress(self, text: str) -> dict:
        """Compress the text provided in parameter

        Args:
            text (str): text to compress
        """
        pass



import os


class File():
    """Class to read / write files
    """
    def __init__(self, path):
        self.path = path

    def read(self):
        """Read file

        Returns:
            str: return text from the film
        """
        with open(self.path) as file:
            return file.read()

    def export(self, frequency, compressedText):
        """Exports frequency and compressed text

        Args:
            frequency (dict): frequency of each characters in initial text
            compressedText (str): compressed text
        """
        dir = os.path.dirname(self.path)
        basename = os.path.splitext(os.path.basename(self.path))[0]

        with open(dir+"/"+basename+"_freq.txt", 'w') as file:
            file.write(str(len(frequency))+"\n")
            frequency = map(
                    lambda x: str(x[0])+" "+str(x[1])+'\n', frequency.items()
                )
            file.writelines(frequency)

        with open(dir+"/"+basename+"_comp.bin", 'wb') as file:
            file.write(compressedText)
